"""This small module implements generic tests for distributions to ensure
that they have been implemented consistently
"""

import os
import numpy as np
from pathlib import Path
import tempfile

from qp.core.ensemble import Ensemble
from qp.plotting import plot_native, plot, plot_pdf_samples_on_axes
from tests.helpers.test_data_helper import NPDF
from qp.core.factory import read, read_metadata, convert
from qp.parameterizations.spline.spline import spline_gen


def assert_all_close(arr, arr2, **kwds):
    """A slightly more informative version of asserting allclose"""
    test_name = kwds.pop("test_name", "test")
    if not np.allclose(arr, arr2, **kwds):  # pragma: no cover
        raise ValueError(
            "%s %.2e %.2e %s"
            % (test_name, (arr - arr2).min(), (arr - arr2).max(), kwds)
        )


def assert_all_small(arr, **kwds):
    """A slightly more informative version of asserting allclose"""
    test_name = kwds.pop("test_name", "test")
    if not np.allclose(arr, 0, **kwds):  # pragma: no cover
        raise ValueError("%s %.2e %.2e %s" % (test_name, arr.min(), arr.max(), kwds))


def build_ensemble(test_data):
    """Build an ensemble from test data in a class"""
    gen_func = test_data["gen_func"]
    ctor_data = test_data["ctor_data"]
    method = test_data.get("method", None)
    try:
        ens = Ensemble(gen_func, data=ctor_data, method=method)
        ancil = test_data.get("ancil")
        if ancil is not None:
            ens.set_ancil(ancil)
        return ens
    except Exception as exep:  # pragma: no cover
        print("Failed to make %s %s %s" % (gen_func, ctor_data, exep))
        raise ValueError from exep


def pdf_func_tests(pdf, test_data, short=False, check_props=True):
    """Run the test for a particular class"""

    xpts = test_data["test_xvals"]
    npdf = NPDF
    if test_data.get("npdf", None) is not None:
        npdf = test_data["npdf"]

    if check_props:
        # if we used the c'tor, make sure the class keeps the data used in the c'tor
        for kv in test_data["ctor_data"].keys():
            test_val = pdf.kwds.get(kv, None)
            if test_val is None:
                if not hasattr(pdf.dist, kv):  # pragma: no cover
                    raise ValueError("%s %s" % (pdf.dist, kv))
                _ = getattr(pdf.dist, kv)

    if hasattr(pdf.dist, "npdf"):
        assert pdf.dist.npdf == pdf.npdf

    assert pdf.shape[0] == npdf

    pdfs = pdf.pdf(xpts)
    if pdf.ndim == 1:
        xslice = xpts[5]
        pdfs_slice = pdf.pdf(xslice)
        check_pdf = pdfs[:, 5].flatten() - pdfs_slice.flatten()
        assert_all_small(check_pdf, atol=2e-2, test_name="pdf")

        xslice = np.expand_dims(xpts[np.arange(pdf.npdf)], -1)
        pdfs_slice = pdf.pdf(xslice)
        pdf_check = np.array([pdfs[i, i] for i in range(pdf.npdf)])
        check_pdfs_slice = pdf_check.flatten() - pdfs_slice.flatten()
        assert_all_small(check_pdfs_slice, atol=2e-2, test_name="pdf_slice")

    if short:  # pragma: no cover
        return pdf

    cdfs = pdf.cdf(xpts)
    quants = np.linspace(0.01, 0.99, 50)

    binw = xpts[1:] - xpts[0:-1]
    if pdf.ndim <= 2:  # changed from == 1
        check_cdf = ((pdfs[:, 0:-1] + pdfs[:, 1:]) * binw / 2).cumsum(axis=-1) - cdfs[
            :, 1:
        ]
    else:
        check_cdf = ((pdfs[:, :, 0:-1] + pdfs[:, :, 1:]) * binw / 2).cumsum(
            axis=-1
        ) - cdfs[:, :, 1:]

    assert_all_small(check_cdf, atol=2e-1, test_name="cdf")

    # FIXME, spline has issues with inversion
    if isinstance(pdf.dist, spline_gen):
        return pdf

    ppfs = pdf.ppf(quants)
    check_cdf_2 = pdf.cdf(ppfs)
    check_ppf = check_cdf_2 - quants
    assert_all_small(check_ppf, atol=2e-2, test_name="ppf")

    if pdf.ndim <= 2:  # changed from ==1
        quants_slice = np.expand_dims(quants[np.arange(pdf.npdf)], -1)
        ppfs_slice = pdf.ppf(quants_slice)
        _ = np.array([ppfs[i, i] for i in range(pdf.npdf)])

        check_ppfs_slice = pdf.cdf(ppfs_slice) - quants_slice
        assert_all_small(check_ppfs_slice, atol=2e-2, test_name="ppf_slice")

    sfs = pdf.sf(xpts)
    check_sf = sfs + cdfs
    assert_all_small(check_sf - 1, atol=1e-5, test_name="sf")

    _ = pdf.isf(quants)
    check_isf = pdf.cdf(ppfs) + quants[::-1]
    assert_all_small(check_isf - 1, atol=5e-2, test_name="isf")
    return pdf


def run_pdf_func_tests(test_class, test_data, short=False, check_props=True):
    """Run the test for a particular class"""

    method = test_data.get("method", None)
    ctor_func = test_class.creation_method(method)
    if ctor_func is None:  # pragma: no cover
        raise KeyError(
            "failed to find creation method %s for class %s" % (method, test_class)
        )
    try:
        pdf = ctor_func(**test_data["ctor_data"])
    except Exception as exep:  # pragma: no cover
        print("Failed to make %s %s %s" % (ctor_func, test_data["ctor_data"], exep))
        raise ValueError from exep

    alloc_kwds = pdf.dist.get_allocation_kwds(pdf.npdf, **test_data["ctor_data"])
    for key, val in alloc_kwds.items():
        assert np.prod(val[0]) == np.size(test_data["ctor_data"][key])

    return pdf_func_tests(pdf, test_data, short=short, check_props=check_props)


def persist_func_test(ensemble, test_data, tmpdir):
    """Run loopback persistence tests on an ensemble"""
    # ftypes = ['fits', 'hdf5', 'pq']
    if ensemble.ndim <= 2:  # changed from ==1
        ftypes = ["fits", "hf5", "h5", "pq", "hdf5"]
    else:
        ftypes = ["fits", "hf5"]

    for ftype in ftypes:
        filekey = test_data.get("filekey", None)
        if filekey is None:
            filekey = ensemble.gen_class.name
        filename = "test_%s.%s" % (filekey, ftype)
        filepath = Path(tmpdir) / filename
        # try:
        #     os.remove(filename)
        # except FileNotFoundError:
        #     pass

        filepath.unlink(missing_ok=True)
        if ftype == "pq":
            filepath.with_stem("test_%sdata.%s" % (filekey, ftype)).unlink(
                missing_ok=True
            )
            filepath.with_stem("test_%smeta.%s" % (filekey, ftype)).unlink(
                missing_ok=True
            )
            filepath.with_stem("test_%sancil.%s" % (filekey, ftype)).unlink(
                missing_ok=True
            )

        if filepath.exists():
            raise RuntimeError(f"{filepath} exists before write")
        if not filepath.parent.exists():
            raise FileNotFoundError(f"{filepath.parent} doesn't exist to write to")
        ensemble.write_to(filepath)
        meta = read_metadata(filepath)
        ens_r = read(filepath)
        meta2 = ens_r.metadata
        # check that reading metadata and main file get same metadata items
        for k, _v in meta.items():
            # we can't actually do a better check than this because the build_tables
            # method in the ensemble class may add extra metadata items and change their type
            assert k in meta2

        diff = ensemble.pdf(test_data["test_xvals"]) - ens_r.pdf(
            test_data["test_xvals"]
        )
        assert_all_small(diff, atol=1e-5, test_name="persist")
        if ensemble.ancil is not None:
            if ftype in ["pq"]:
                continue
            diff2 = ensemble.ancil["zmode"] - ens_r.ancil["zmode"]
            assert_all_small(diff2, atol=1e-5, test_name="persist_ancil")

        # TODO: figure out if I need to delete files here
        # try:
        #     pass
        # except FileNotFoundError:
        #     pass
        filepath.unlink(missing_ok=True)
        filepath.with_stem("test_%sdata.%s" % (filekey, ftype)).unlink(missing_ok=True)
        filepath.with_stem("test_%smeta.%s" % (filekey, ftype)).unlink(missing_ok=True)
        filepath.with_stem("test_%sancil.%s" % (filekey, ftype)).unlink(missing_ok=True)
        # os.unlink(filepath.replace(".%s" % ftype, "data.%s" % ftype))
        # os.unlink(filepath.replace(".%s" % ftype, "meta.%s" % ftype))
        # except FileNotFoundError:
        #     pass


def run_persist_func_tests(test_data):
    """Run the test for a particular class"""
    ens = build_ensemble(test_data)
    tmpdir = tempfile.TemporaryDirectory()
    persist_func_test(ens, test_data, tmpdir.name)
    tmpdir.cleanup()


def run_convert_tests(ens_orig, gen_class, test_data, **kwargs):
    """Run the test for a particular class"""

    xpts = test_data["test_xvals"]

    binw = np.mean(xpts[1:] - xpts[0:-1])
    atol = kwargs.get("atol_diff", 1e-2) / binw
    atol2 = kwargs.get("atol_diff2", 1e-2) / binw

    ens1 = ens_orig.convert_to(gen_class, **test_data["convert_data"])
    diffs = ens_orig.pdf(xpts) - ens1.pdf(xpts)
    assert_all_small(diffs, atol=atol, test_name="convert")

    ens2 = convert(ens_orig, gen_class.name, **test_data["convert_data"])
    diffs = ens_orig.pdf(xpts) - ens2.pdf(xpts)
    assert_all_small(diffs, atol=atol2, test_name="convert2")

    assert ens_orig.shape[:-1] == ens1.shape[:-1]
    assert ens_orig.shape[:-1] == ens2.shape[:-1]
    assert ens_orig.frozen.shape[:-1] == ens2.shape[:-1]
    if hasattr(ens2.dist, "shape"):
        assert ens2.dist.shape[:-1] == ens2.shape[:-1]


def plotting_func_tests(ensemble, do_samples=False):
    """Run the test for a particular class"""
    pdf = ensemble[0]
    fig, axes = plot_native(ensemble, xlim=(-5, 5))
    assert fig is not None
    assert axes is not None
    fig, axes = plot(ensemble, axes=axes)
    assert fig is not None
    assert axes is not None
    fig, axes = plot_native(pdf.frozen, xlim=(-5, 5))
    assert fig is not None
    assert axes is not None
    fig, axes = plot(pdf.frozen, xlim=(-5, 5))
    assert fig is not None
    assert axes is not None
    if do_samples:
        samples = pdf.rvs(size=1000)
        plot_pdf_samples_on_axes(axes, pdf, samples)


def run_plotting_func_tests(test_data, do_samples=False):
    """Run the test for a particular class"""
    ens = build_ensemble(test_data)
    if ens.ndim > 2:  # changed from != 1
        return
    plotting_func_tests(ens, do_samples=do_samples)



from __future__ import annotations

from typing import Any

import numpy as np
from scipy import interpolate as sciinterp

from .sparse_rep import build_sparse_representation


def extract_sparse_from_xy(
    in_dist: "Ensemble", **kwargs
) -> dict[str, Any]:  # pragma: no cover
    """Extract sparse representation from an xy interpolated representation

    Parameters
    ----------
    in_dist : Ensemble
        Input distributions

    Other Parameters
    ----------------
    xvals : ArrayLike
        Used to override the y-values
    xvals : ArrayLike
        Used to override the x-values
    nvals : int
        Used to override the number of bins

    Returns
    -------
    metadata : dict[str, Any]
        Dictionary with data for sparse representation

    Notes
    -----
    This function will rebin to a grid more suited to the in_dist support by
    removing x-values corrsponding to y=0
    """
    default = in_dist.objdata["yvals"]
    yvals = kwargs.pop("yvals", default)
    default = in_dist.metadata["xvals"][0]
    xvals = kwargs.pop("xvals", default)
    nvals = kwargs.pop("nvals", 300)
    # rebin to a grid more suited to the in_dist support
    xmin = np.min(xvals)
    _, j = np.where(yvals > 0)
    xmax = np.max(xvals[j])
    newx = np.linspace(xmin, xmax, nvals)
    interp = sciinterp.interp1d(xvals, yvals, assume_sorted=True)
    newpdf = interp(newx)
    sparse_indices, metadata, _ = build_sparse_representation(newx, newpdf)
    metadata["xvals"] = newx
    metadata["sparse_indices"] = sparse_indices
    metadata.pop("Ntot")
    return metadata



""" Lazy loading modules """

from tables_io.lazy_modules import lazyImport

mpl = lazyImport("matplotlib")
plt = lazyImport("matplotlib.pyplot")
mixture = lazyImport("sklearn.mixture")
pytdigest = lazyImport("pytdigest")



from __future__ import annotations

import numpy as np
from numpy.typing import ArrayLike
from scipy import stats as sps
from scipy.integrate import quad
from scipy.interpolate import splev, splrep
from scipy.special import errstate  # pylint: disable=no-name-in-module

from ...utils.conversion import extract_xy_vals


def normalize_spline(
    xvals: ArrayLike, yvals: ArrayLike, limits: tuple[float, float], **kwargs
) -> np.ndarray[float]:
    """
    Normalize a set of 1D interpolators

    Parameters
    ----------
    xvals : ArrayLike
        X-values used for the spline, should be a 2D array.
    yvals : ArrayLike
        Y-values used for the spline, should be a 2D array.
    limits : tuple[float, float]
        Lower and Upper limits of integration
    kwargs :
        Passed to the `scipy.integrate.quad` integration function

    Returns
    -------
    ynorm : np.ndarray[float]
        Normalized y-vals
    """

    def row_integral(irow):
        def spl(xv):
            return splev(xv, splrep(xvals[irow], yvals[irow]))

        return quad(spl, limits[0], limits[1], **kwargs)[0]

    vv = np.vectorize(row_integral)
    with errstate(all="ignore"):
        integrals = vv(np.arange(xvals.shape[0]))
    return (yvals.T / integrals).T


def build_splines(
    xvals: ArrayLike, yvals: ArrayLike
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Build a set of 1D spline representations

    Parameters
    ----------
    xvals : ArrayLike
        X-values used for the spline
    yvals : ArrayLike
        Y-values used for the spline

    Returns
    -------
    splx : np.ndarray
        Spline knot xvalues
    sply : np.ndarray
        Spline knot yvalues
    spln : np.ndarray
        Spline knot order parameters
    """
    l_x = []
    l_y = []
    l_n = []
    for xrow, yrow in zip(xvals, yvals):
        rep = splrep(xrow, yrow)
        l_x.append(rep[0])
        l_y.append(rep[1])
        l_n.append(rep[2])
    return np.vstack(l_x), np.vstack(l_y), np.vstack(l_n)


# Conversion utility functions


def spline_extract_xy_vals(in_dist: "Ensemble", **kwargs) -> dict[str, Any]:
    """Wrapper for extract_xy_vals. Convert using a set of x and y values.

    Parameters
    ----------
    in_dist : Ensemble
        Input distributions
    xvals : ArrayLike
        Locations at which the pdf is evaluated

    Returns
    -------
    data : dict[str, Any]
        The extracted data
    """

    xvals = kwargs.pop("xvals", None)
    if xvals is None:  # pragma: no cover
        raise ValueError("To convert using extract_xy_vals you must specify xvals")
    return extract_xy_vals(in_dist, xvals)


def extract_samples(in_dist: "Ensemble", **kwargs) -> dict[str, np.ndarray | None]:
    """Convert using a set of values sampled from the PDF

    Parameters
    ----------
    in_dist : Ensemble
        Input distributions

    Other Parameters
    ----------------
    size : int
        Number of samples to generate

    Returns
    -------
    data : dict[str, np.ndarray | None]
        The extracted data
    """
    samples = in_dist.rvs(size=kwargs.pop("size", 1000))
    xvals = kwargs.pop("xvals")
    return dict(samples=samples, xvals=xvals, yvals=None)


# Creation utility functions


def build_kdes(samples: ArrayLike, **kwargs) -> list[sps.gaussian_kde]:
    """
    Build a set of Gaussian Kernel Density Estimates

    Parameters
    ----------
    samples : ArrayLike
        X-values used for the spline
    kwargs
        Passed to the `scipy.stats.gaussian_kde` constructor

    Returns
    -------
    kdes : list[sps.gaussian_kde]
    """
    return [sps.gaussian_kde(row, **kwargs) for row in samples]


def evaluate_kdes(xvals: ArrayLike, kdes: list[sps.gaussian_kde]) -> np.ndarray:
    """
    Build a evaluate a set of kdes

    Parameters
    ----------
    xvals : ArrayLike
        X-values used for the spline
    kdes : list[sps.gaussian_kde]
        The kernel density estimates

    Returns
    -------
    yvals : np.ndarray
        The kdes evaluated at the xvals
    """
    return np.vstack([kde(xvals) for kde in kdes])



"""Utility functions for array handling in the the qp package"""

from __future__ import annotations

import sys

import numpy as np
from typing import Mapping, Union
from numpy.typing import ArrayLike

# epsilon = sys.float_info.epsilon
# infty = sys.float_info.max * epsilon
# lims = (epsilon, 1.0)

CASE_PRODUCT = 0
CASE_FACTOR = 1
CASE_2D = 2
CASE_FLAT = 3


_ = """
def normalize_quantiles(in_data, threshold=epsilon, vb=False):
    Evaluates PDF from quantiles including endpoints from linear extrapolation

    Parameters
    ----------
    in_data: tuple, numpy.ndarray, float
        tuple of CDF values iy corresponding to quantiles and the points x at
        which those CDF values are achieved
    threshold: float, optional
        optional minimum threshold for PDF
    vb: boolean, optional
        be careful and print progress to stdout?

    Returns
    -------
    out_data: tuple, ndarray, float
        tuple of values x at which CDF is achieved, including extrema, and
        normalized PDF values y at x

    (iy, x) = in_data
    (xs, ys) = evaluate_quantiles((iy, x), vb=vb)
    # xs = xs[1:-1]
    # ys = ys[1:-1]
    x_min = xs[0] - 2 * iy[0] / ys[0]
    x_max = xs[-1] + 2 * (1. - iy[-1]) / ys[-1]
    xs = sandwich(xs, (x_min, x_max))
    ys = sandwich(ys, (threshold, threshold))
    out_data = (xs, ys)
    return out_data

"""


def edge_to_center(edges: ArrayLike) -> np.ndarray:
    """Return the centers of a set of bins given the edges"""
    return 0.5 * (edges[1:] + edges[:-1])


def bin_widths(edges: ArrayLike) -> np.ndarray:
    """Return the widths of a set of bins given the edges"""
    return edges[1:] - edges[:-1]


def get_bin_indices(bins: ArrayLike, x: ArrayLike) -> np.ndarray[int]:
    """Return the bin indexes for a set of values

    If the bins are equal width this will use arithmetic,
    If the bins are not equal width this will use a binary search
    """
    widths = bin_widths(bins)
    n_bins = np.size(bins) - 1
    if np.allclose(widths, widths[0]):
        idx = np.atleast_1d(np.floor((x - bins[0]) / widths[0]).astype(int))
    else:
        idx = np.atleast_1d(np.searchsorted(bins, x, side="left") - 1)
    mask = (idx >= 0) * (idx < bins.size - 1)
    np.putmask(idx, 1 - mask, 0)
    xshape = np.shape(x)
    return idx.reshape(xshape).clip(0, n_bins - 1), mask.reshape(xshape)


def get_eval_case(x: ArrayLike, row: ArrayLike) -> tuple[int, np.ndarray, np.ndarray]:
    """Figure out which of the various input formats scipy.stats has passed us

    Parameters
    ----------
    x : ArrayLike
        Pdf x-vals
    row : ArrayLike
        Pdf row indices

    Returns
    -------
    case : int
        The case code
    xx : np.ndarray
        The x-values properly shaped
    rr : np.ndarrray
        The y-values, properly shaped

    Notes
    -----
    The cases are:

    CASE_FLAT : x, row have shapes (n), (n) and do not factor
    CASE_FACTOR : x, row have shapes (n), (n) but can be factored to shapes (1, nx) and (npdf, 1)
                  (i.e., they were flattened by scipy)
    CASE_PRODUCT : x, row have shapes (1, nx) and (npdf, 1)
    CASE_2D : x, row have shapes (npdf, nx) and (npdf, nx)

    """
    nd_x = np.ndim(x)
    nd_row = np.ndim(row)
    # if nd_x > 2 or nd_row > 2:  #pragma: no cover
    #    raise ValueError("Too many dimensions: x(%s), row(%s)" % (np.shape(x), np.shape(row)))
    if nd_x >= 2 and nd_row != 1:
        return CASE_2D, x, row
    if nd_x >= 2 and nd_row == 1:  # pragma: no cover
        raise ValueError(
            "Dimension mismatch: x(%s), row(%s)" % (np.shape(x), np.shape(row))
        )
    if nd_row >= 2:
        return CASE_PRODUCT, x, row
    if np.size(x) == 1 or np.size(row) == 1:
        return CASE_FLAT, x, row
    xx = np.unique(x)
    rr = np.unique(row)
    if np.size(xx) == np.size(x):
        xx = x
    if np.size(rr) == np.size(row):
        rr = row
    if np.size(xx) * np.size(rr) != np.size(x):
        return CASE_FLAT, x, row
    return CASE_FACTOR, xx, np.expand_dims(rr, -1)


def profile(
    x_data: ArrayLike, y_data: ArrayLike, x_bins: ArrayLike, std: bool = True
) -> tuple[np.ndarray[float], np.ndarray[float]]:
    """Make a 'profile' plot

    Parameters
    ----------
    x_data : ArrayLike, length n
        The x-values
    y_data : ArrayLike, length n
        The y-values
    x_bins : ArrayLike, length nbins+1
        The values of the bin edges
    std : bool, optional
        If true, return the standard deviations, if false return the errors on the
        means, default True.

    Returns
    -------
    vals : np.ndarray[float], length nbins
        The means
    errs : np.ndarray[float], length nbins
        The standard deviations or errors on the means
    """
    idx, mask = get_bin_indices(x_bins, x_data)
    count = np.zeros(x_bins.size - 1)
    vals = np.zeros(x_bins.size - 1)
    errs = np.zeros(x_bins.size - 1)
    for i in range(x_bins.size - 1):
        mask_col = mask * (idx == i)
        count[i] = mask_col.sum()
        if mask_col.sum() == 0:  # pragma: no cover
            vals[i] = np.nan
            errs[i] = np.nan
            continue
        masked_vals = y_data[mask_col]
        vals[i] = masked_vals.mean()
        errs[i] = masked_vals.std()
    if not std:
        errs /= np.sqrt(count)
    return vals, errs


def reshape_to_pdf_size(vals: np.ndarray, split_dim: int) -> np.ndarray:
    """Reshape an array to match the number of PDFs in a distribution

    Parameters
    ----------
    vals : np.ndarray
        The input array
    split_dim : int
        The dimension at which to split between pdf indices and per_pdf indices

    Returns
    -------
    out : np.ndarray
        The reshaped array
    """
    in_shape = np.shape(vals)
    npdf = np.prod(in_shape[:split_dim]).astype(int)
    per_pdf = in_shape[split_dim:]
    out_shape = np.hstack([npdf, per_pdf])
    return vals.reshape(out_shape)


def reshape_to_pdf_shape(
    vals: np.ndarray, pdf_shape: int, per_pdf: int | ArrayLike
) -> np.ndarray:
    """Reshape an array to match the shape of PDFs in a distribution

    Parameters
    ----------
    vals : np.ndarray
        The input array
    pdf_shape : int
        The shape for the pdfs
    per_pdf : int | ArrayLike
        The shape per pdf

    Returns
    -------
    out : np.ndarray
        The reshaped array
    """
    outshape = np.hstack([pdf_shape, per_pdf])
    return vals.reshape(outshape)


def encode_strings(data: Mapping[str, np.ndarray]) -> Mapping[str, np.ndarray]:
    """Encodes any dictionary values that are Unicode strings (or just strings
    if not numpy arrays). Other data types are not affected.

    Parameters
    ----------
    data : Mapping[str, np.ndarray]
        Dictionary of data to encode.

    Returns
    -------
    Mapping[str, np.ndarray]
        Dictionary of data with strings encoded.
    """

    converted_data = {}
    for key, val in data.items():
        new_val = val
        if isinstance(val, np.ndarray):
            # encode unicode strings as bytes to work with hdf5
            if val.dtype.kind == "U":
                new_val = np.strings.encode(val, "utf-8")
        else:
            # is not a numpy array
            if isinstance(val[0], str):
                new_val = np.strings.encode(val, "utf-8")

        converted_data[key] = new_val

    return converted_data


def decode_strings(data: Mapping[str, np.ndarray]) -> Mapping[str, np.ndarray]:
    """Decodes dictionary values that have been encoded (dtype = bytes). Other
    data types are not affected.

    Parameters
    ----------
    data : Mapping[str, np.ndarray]
        The dictionary of data to be decoded.

    Returns
    -------
    Mapping[str, np.ndarray]
        The dictionary of data with any strings decoded.
    """
    converted_data = {}
    for key, val in data.items():
        new_val = val

        if isinstance(val, np.ndarray):
            # decode any string objects as necessary
            if val.dtype.kind == "S":
                new_val = np.strings.decode(val, "utf-8")
        else:
            # decode string objects that are not numpy arrays
            if isinstance(val[0], bytes):
                new_val = np.strings.decode(val, "utf-8")

        converted_data[key] = new_val

    return converted_data


def reduce_dimensions(arr: np.ndarray, x: ArrayLike) -> Union[float, np.ndarray]:
    """If the given array has dimensionality greater than x, reduces its dimensionality
    to match x, if this will not result in a loss of data.

    Parameters
    ----------
    arr : np.ndarray
        Array to reduce dimensionality
    x : ArrayLike
        Object to match dimensionality to

    Returns
    -------
    Union[float, np.ndarray]
        The array with dimension reduced (if possible)
    """
    if np.ndim(x) < 1 and np.ndim(arr) >= 2:
        if np.shape(arr) == (1, 1):
            return arr.item()
    if np.ndim(x) == 1 and np.ndim(arr) > 1:
        if arr.shape[0] == 1:
            return np.squeeze(arr)
    return arr



import logging

import numpy as np
from scipy import stats

import qp
from .array_metrics import quick_anderson_ksamp
from .metrics import calculate_outlier_rate

DEFAULT_QUANTS = np.linspace(0, 1, 100)


class PIT:
    """PIT(qp_ens, true_vals, eval_grid=DEFAULT_QUANTS)
    Probability Integral Transform

    Parameters
    ----------
    qp_ens : Ensemble
        A collection of N distribution objects
    true_vals : [float]
        An array-like sequence of N float values representing the known true value for each distribution
    eval_grid : [float], optional
        A strictly increasing array-like sequence in the range [0,1], by default DEFAULT_QUANTS

    Returns
    -------
    PIT object
        An object with an Ensemble containing the PIT distribution, and a full set of PIT samples.
    """

    def __init__(self, qp_ens, true_vals, eval_grid=DEFAULT_QUANTS):
        """We will create a quantile Ensemble to store the PIT distribution, but also store the
        full set of PIT samples as ancillary data of the (single PDF) ensemble.

        Parameters
        ----------
        qp_ens : Ensemble
            A collection of N distribution objects
        true_vals : [float]
            An array-like sequence of N float values representing the known true value for each distribution
        eval_grid : [float], optional
            A strictly increasing array-like sequence in the range [0,1], by default DEFAULT_QUANTS
        """

        self._pit_samps = self._gather_pit_samples(qp_ens, true_vals)

        n_pit = np.min([len(self._pit_samps), len(eval_grid)])
        if n_pit < len(eval_grid):
            logging.warning(
                "Number of pit samples is smaller than the evaluation grid size. "
                "Will create a new evaluation grid with size = number of pit samples"
            )
            eval_grid = np.linspace(0, 1, n_pit)

        data_quants = np.quantile(self._pit_samps, eval_grid)

        self._pit = self._produce_output_ensemble(data_quants, eval_grid)

    @property
    def pit_samps(self):
        """Returns the PIT samples. i.e. ``CDF(true_vals)`` for each distribution
        in the Ensemble used to initialize the PIT object.

        Returns
        -------
        np.array
            An array of floats
        """
        return self._pit_samps

    @property
    def pit(self):
        """Return the PIT Ensemble object

        Returns
        -------
        qp.Ensemble
            An Ensemble containing 1 qp.quant distribution.
        """
        return self._pit

    def calculate_pit_meta_metrics(self):
        """Convenience method that will calculate all of the PIT meta metrics and return
        them as a dictionary.

        Returns
        -------
        dictionary
            The collection of PIT statistics
        """
        pit_meta_metrics = {}

        pit_meta_metrics["ad"] = self.evaluate_PIT_anderson_ksamp()
        pit_meta_metrics["cvm"] = self.evaluate_PIT_CvM()
        pit_meta_metrics["ks"] = self.evaluate_PIT_KS()
        pit_meta_metrics["outlier_rate"] = self.evaluate_PIT_outlier_rate()

        return pit_meta_metrics

    def evaluate_PIT_anderson_ksamp(self, pit_min=0.0, pit_max=1.0):
        """Use scipy.stats.anderson_ksamp to compute the Anderson-Darling statistic
        for the cdf(truth) values by comparing with a uniform distribution between 0 and 1.
        Up to the current version (1.9.3), scipy.stats.anderson does not support
        uniform distributions as reference for 1-sample test, therefore we create a uniform
        "distribution" and pass it as the second value in the list of parameters to the scipy
        implementation of k-sample Anderson-Darling.
        For details see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.anderson_ksamp.html

        Parameters
        ----------
        pit_min : float, optional
            Minimum PIT value to accept, by default 0.
        pit_max : float, optional
            Maximum PIT value to accept, by default 1.

        Returns
        -------
        array
            A array of objects with attributes `statistic`, `critical_values`, and `significance_level`.
            For details see:
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.anderson_ksamp.html
        """
        # Removed the CDF values that are outside the min/max range
        trimmed_pit_values = self._trim_pit_values(pit_min, pit_max)

        uniform_yvals = np.linspace(pit_min, pit_max, len(trimmed_pit_values))

        return quick_anderson_ksamp(trimmed_pit_values, uniform_yvals)

    def evaluate_PIT_CvM(self):
        """Calculate the Cramer von Mises statistic using scipy.stats.cramervonmises using self._pit_samps
        compared to a uniform distribution. For more details see:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.cramervonmises.html

        Returns
        -------
        array
            A array of objects with attributes `statistic` and `pvalue`
            For details see:
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.cramervonmises.html
        """
        return stats.cramervonmises(self._pit_samps, stats.uniform.cdf)

    def evaluate_PIT_KS(self):
        """Calculate the Kolmogorov-Smirnov statistic using scipy.stats.kstest. For more details see:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.kstest.html

        Returns
        -------
        array
            A array of objects with attributes `statistic` and `pvalue`.
            For details see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.kstest.html
        """
        return stats.kstest(self._pit_samps, stats.uniform.cdf)

    def evaluate_PIT_outlier_rate(self, pit_min=0.0001, pit_max=0.9999):
        """Compute fraction of PIT outliers by evaluating the CDF of the distribution in the PIT Ensemble
        at `pit_min` and `pit_max`.

        Parameters
        ----------
        pit_min : float, optional
            Lower bound for outliers, by default 0.0001
        pit_max : float, optional
            Upper bound for outliers, by default 0.9999

        Returns
        -------
        float
            The percentage of outliers in this distribution given the min and max bounds.
        """
        return calculate_outlier_rate(self._pit, pit_min, pit_max)[0]

    @classmethod
    def _gather_pit_samples(cls, qp_ens, true_vals):
        pit_samples = np.squeeze(qp_ens.cdf(np.vstack(true_vals)))

        # These two lines set all `NaN` values to 0. This may or may not make sense
        # Alternatively if it's better to simply remove the `NaN`, this can be done
        # efficiently on line 61 with `data_quants = np.nanquantile(...)`.`
        sample_mask = np.isfinite(pit_samples)
        pit_samples[~sample_mask] = 0
        if not np.all(sample_mask):  # pragma: no cover
            logging.warning(
                "Some PIT samples were `NaN`. They have been replacd with 0."
            )

        return pit_samples

    @classmethod
    def _produce_output_ensemble(cls, data_quants, eval_grid):
        # Remove duplicates values as well as values outside the range (0,1)
        _, unique_indices = np.unique(data_quants, return_index=True)
        unique_data_quants = data_quants[unique_indices]
        unique_eval_grid = eval_grid[unique_indices]
        quant_mask = cls._create_quant_mask(unique_data_quants)

        return qp.Ensemble(
            qp.quant,
            data=dict(
                quants=unique_eval_grid[quant_mask],
                locs=np.atleast_2d(unique_data_quants[quant_mask]),
            ),
        )

    @classmethod
    def _create_quant_mask(cls, data_quants):
        """Create a numpy mask such that, when applied only values greater than
        0 and less than 1.0 are kept. While this function is fairly simple,
        separating it into a small helper method makes testing much easier.

        Parameters
        ----------
        data_quants : np.array [float]
            An array of values.

        Returns
        -------
        np.array [bool]
            The boolean mask
        """

        return np.bitwise_and(data_quants > 0.0, data_quants < 1)

    def _trim_pit_values(self, cdf_min, cdf_max):
        """Remove and report any cdf(x) that are outside the min/max range.

        Parameters
        ----------
        cdf_min : float
            The minimum cdf(x) value to accept
        cdf_max : float
            The maximum cdf(x) value to accept

        Returns
        -------
        clean: [float]
            The list of PIT values within the min/max range.
        """
        # Create truth mask for pit values between cdf_min and pit max
        mask = (self._pit_samps >= cdf_min) & (self._pit_samps <= cdf_max)

        # Keep pit values that are within the min/max range
        pits_clean = self._pit_samps[mask]

        # Determine how many pit values were dropped and warn the user.
        diff = len(self._pit_samps) - len(pits_clean)
        if diff > 0:
            logging.warning("Removed %d PITs from the sample.", diff)

        return pits_clean
